
{-# LANGUAGE CPP #-}
{-# LANGUAGE DeriveDataTypeable #-}

--------------------------------------------------------------------------------
-- |
-- Module      :  Data.Stream
-- Copyright   :  (c) Wouter Swierstra <wss@cs.nott.ac.uk>, Bas van Dijk <v.dijk.bas@gmail.com>
-- License     :  BSD3
--
-- Maintainer  :  Bas van Dijk <v.dijk.bas@gmail.com>
--
-- Streams are infinite lists. Most operations on streams are
-- completely analogous to the definition in Data.List.

--------------------------------------------------------------------------------


--------------------------------------------------------------------------------
-- Exports
--------------------------------------------------------------------------------

module Data.Stream
    (
    -- * The type of streams
      Stream(..)

    -- * Basic functions
    , fold
    , head
    , tail
    , inits
    , tails

    -- * Stream transformations
    , map
    , mapAccum, mapAccum'
    , concat
    , concatMap
    , intersperse
    , intercalate
    , transpose
    , subsequences
    , permutations
    , interleave
    , scan, scan'
    , scan1, scan1'

    -- * Diagonals
    , diagonal
    , diagonals
    , unDiagonals
    , concatDiagonals
    , unConcatDiagonals

    -- * Building streams
    , iterate, iterate'
    , repeat
    , replicateCont
    , cycle
    , unfold, unfold'

    -- * Extracting sublists
    , take,    genericTake
    , drop,    genericDrop
    , splitAt, genericSplitAt
    , takeWhile
    , dropWhile
    , span
    , break
    , filter
    , partition
    , group
    , groupBy

    -- * Stream prefixes
    , isPrefixOf
    , stripPrefix

    -- * Indexing streams
    , (!!), genericIndex
    , elemIndex
    , elemIndices
    , findIndex
    , findIndices

    -- * Searching streams
    , lookup

    -- * Zipping and unzipping streams
    , zip,     zip3,     zip4,     zip5,     zip6,     zip7
    , zipWith, zipWith3, zipWith4, zipWith5, zipWith6, zipWith7
    , unzip,   unzip3,   unzip4,   unzip5,   unzip6,   unzip7

    -- * Functions on streams of characters
    , words
    , unwords
    , lines
    , unlines

    -- * Converting to and from an lists
    , toList
    , fromList
    , fromListCont

     -- * Set operations
    , nub
    , nubBy
    , delete
    , deleteBy

     -- * Ordered streams
    , insert
    , insertBy
    , merge
    , mergeBy
    )
    where


--------------------------------------------------------------------------------
-- Imports
--------------------------------------------------------------------------------

import Prelude hiding (head, tail, map,
  iterate, take, drop, takeWhile,
  dropWhile, repeat, cycle, filter, (!!),
  zip, zip3, unzip, unzip3, zipWith3, zipWith,
  words,unwords,lines,unlines, break, span, splitAt,
  concat, concatMap, or, any, lookup)

import qualified Prelude

import Control.Monad       (liftM2)
import Control.Applicative (Applicative, pure, (<*>), liftA2)
import Control.Monad.Fix   (MonadFix, mfix, fix)
import Control.Arrow       (first, second)
import Data.Char           (isSpace)
import Data.Typeable       (Typeable)
import Data.Monoid         (mappend)
import Data.Traversable    (Traversable, traverse, sequenceA)
import Data.Foldable       (Foldable, foldMap)
import qualified Data.Foldable

#ifdef BASE_4
import Data.Data (Data)
#else
import Data.Generics (Data)
#endif

import Test.QuickCheck ( Arbitrary,   arbitrary
                       , CoArbitrary, coarbitrary
                       )
import Test.LazySmallCheck (Serial, series, cons2)


--------------------------------------------------------------------------------
-- Type
--------------------------------------------------------------------------------

-- | An infinite sequence.
--
-- /Beware/: If you use any function from the @ Eq @ or @ Ord @ class
-- two compare to /equal/ streams, these functions will diverge.

data Stream a = a ::: Stream a -- ^ The symbol (@:::@) is chosen to
                               -- resemble the list constructor
                               -- (@:@). We use three colons to
                               -- indicate that this is a \"very big\"
                               -- list.
                deriving (Eq, Ord, Typeable, Data)

infixr 5 :::


--------------------------------------------------------------------------------
-- Instances
--------------------------------------------------------------------------------

instance Functor Stream where
  fmap = map

instance Applicative Stream where
  pure  = repeat
  (<*>) = zipWith ($)

instance Monad Stream where
  return   = repeat
  xs >>= f = join $ map f xs
    where
      join = diagonal

-- | 'diagonal' @xss@ returns the main diagonal of the stream of
-- streams @xss@ that runs from upper left to lower right. For
-- example:
--
-- @ diagonal
-- (   (1 ::: _ ::: _ ::: _ ::: ...)
-- ::: (_ ::: 2 ::: _ ::: _ ::: ...)
-- ::: (_ ::: _ ::: 3 ::: _ ::: ...)
-- ::: (_ ::: _ ::: _ ::: 4 ::: ...)
-- ...
-- ) == 1 ::: 2 ::: 3 ::: 4 ::: ...
-- @
--
-- Note that 'diagonal' @=@ @join@.
diagonal :: Stream (Stream a) -> Stream a
diagonal (xs ::: xss) = head xs ::: diagonal (map tail xss)

instance MonadFix Stream where
    mfix f = case fix (f . head) of
               (x ::: _) -> x ::: mfix (tail . f)

instance Foldable Stream where
    fold      = fold mappend
    foldMap f = fold (mappend . f)

instance Traversable Stream where
    sequenceA  = fold (liftA2 (:::))
    traverse f = fold (liftA2 (:::) . f)

-- Parsing & Printing ----------------------------------------------------------

consPrecedence :: Int
consPrecedence = 5

-- | A Show instance for Streams that takes the right associativity into
-- account and so doesn't put parenthesis around the tail of the Stream.
-- Note that 'show' returns an infinite 'String'.
instance Show a => Show (Stream a) where
    showsPrec p (x ::: xs) = showParen (p > consPrecedence)   $
                             showsPrec (consPrecedence + 1) x .
                             showString " ::: "               .
                             showsPrec consPrecedence xs

-- TODO: Think about if it's possible to define a Read instance for
-- Streams.
--
-- I would like the following to return [1,2,3,1,2,3,1,2,3,1]:
--
-- testReadShow :: [Int]
-- testReadShow = take 10 $ read $ show $ cycle [1,2,3]
--
-- However with the following Read instance it loops:
--
-- instance Read a => Read (Stream a) where
--     readsPrec p = readParen (p > consPrecedence) $
--                   \r -> [ (x ::: xs, w)
--                         | (x, s)     <- readsPrec (consPrecedence + 1) r
--                         , (":::", t) <- lex s
--                         , (xs, w)    <- readsPrec consPrecedence t
--                         ]
--
-- The problem is that 'show $ cycle [1,2,3]' is an infinite String
-- and 'read' wants to parse the whole input before returning a
-- result.
--

-- I think that as long as there aren't parenthesis around the tail of
-- a Stream (as in: "1 ::: 2 ::: 3 ::: 4 ::: ...") it should be
-- possible to define some kind of "online" parser. Maybe we can use
-- the "Polish Parsers" from Doaitse Swierstra et al..


-- Test data generation  -------------------------------------------------------

instance Arbitrary a => Arbitrary (Stream a) where
  arbitrary = liftM2 (:::) arbitrary arbitrary

instance CoArbitrary a => CoArbitrary (Stream a) where
  coarbitrary xs gen = do n <- arbitrary
                          coarbitrary (take (abs n) xs) gen

instance Serial a => Serial (Stream a) where
    series = cons2 (:::)


--------------------------------------------------------------------------------
-- Basic functions
--------------------------------------------------------------------------------

-- | Fold a stream using a binary operator.
--
-- /Beware/: 'fold' @f@ is only guaranteed to be safe (won't diverge)
-- if @f@ is a /productive/ function, i.e. if it directly returns a
-- value in Weak Head Normal Form (WHNF). A value in WHNF is either a
-- function or a constructor applied to zero or more arguments. Such a
-- value is safe because it won't be reduced directly.
fold :: (a -> b -> b) -> Stream a -> b
fold f = fold_f
    where
      fold_f (x ::: xs) = x `f` fold_f xs

-- | Extract the first element of the stream.
head :: Stream a -> a
head (x ::: _) = x

-- | Extract the stream following the head of the stream.
tail :: Stream a -> Stream a
tail (_ ::: xs) = xs

-- | The 'inits' function takes a stream @xs@ and returns all the
-- finite prefixes of @xs@.
--
-- Note that this 'inits' is lazier then @Data.List.inits@! Formally:
--
-- > inits _|_ = [] ::: _|_
--
-- while @Data.List.inits@:
--
-- > inits _|_ = _|_
inits :: Stream a -> Stream [a]
inits xxs = [] ::: case xxs of
                     x ::: xs -> map (x:) (inits xs)

-- | The 'tails' function takes a stream @xs@ and returns all the
-- suffixes of @xs@.
tails :: Stream a -> Stream (Stream a)
tails xs = xs ::: tails (tail xs)


--------------------------------------------------------------------------------
-- Stream transformations
--------------------------------------------------------------------------------

-- | Apply a function uniformly over all the elements of a stream.
map :: (a -> b) -> Stream a -> Stream b
map f = fold ((:::) . f)

-- | 'mapAccum' behaves like a combination of 'map' and 'fold'; it
-- applies a function to each element of the stream while passing an
-- accumulating parameter.
mapAccum :: (acc -> a -> (acc, b)) -> acc -> Stream a -> Stream b
mapAccum f = mapAccum_f
    where
      mapAccum_f s (x ::: xs) = let (s', y) = f s x
                                in y ::: mapAccum_f s' xs

-- | A strict 'mapAccum'. For example:
--
-- > mapAccum (\z x -> (z+1, z)) 0 (repeat 1) !! 1000000
-- > *** Exception: stack overflow
--
-- > mapAccum' (\z x -> (z+1, z)) 0 (repeat 1) !! 1000000
-- > 1000000
mapAccum' :: (acc -> a -> (acc, b)) -> acc -> Stream a -> Stream b
mapAccum' f = mapAccum'_f
    where
      mapAccum'_f s (x ::: xs) = let (s', y) = f s x
                                 in y ::: (mapAccum'_f $! s') xs

-- | Concatenate a stream of lists.
concat :: Stream [a] -> Stream a
concat = fold $ flip $ foldr (:::)

-- | 'concatMap' @f@ @xs@ is equivalent to @concat $ map f xs@. So it
-- | maps a, list producing, function over a stream and concatenate
-- | the results.
concatMap :: (a -> [b]) -> Stream a -> Stream b
concatMap f = fold $ (flip $ foldr (:::)) . f

-- | The intersperse function takes an element and a stream and
-- `intersperses' that element between the elements of the stream.
intersperse :: a -> Stream a -> Stream a
intersperse y = fold $ \x ->  (x :::) . (y :::)

-- | 'intercalate' @xs@ @xss@ is equivalent to @concat $ intersperse xs xss@.
-- It inserts the list @xs@ in between the lists in @xss@ and concatenates the result
intercalate :: [a] -> Stream [a] -> Stream a
intercalate xs = fold $ \ys -> fromListCont ys . fromListCont xs

-- | The 'transpose' function transposes the rows and columns of its argument.
-- For example:
--
-- @
-- transpose (   (0 ::: 1 ::: 3  ::: ...)
--           ::: (2 ::: 4 ::: 7  ::: ...)
--           ::: (5 ::: 8 ::: 12 ::: ...)
--           ::: ...
--           ) ==
--               (0 ::: 2 ::: 5  ::: ...)
--           ::: (1 ::: 4 ::: 8  ::: ...)
--           ::: (3 ::: 7 ::: 12 ::: ...)
--           ::: ...
-- @
transpose :: Stream (Stream a) -> Stream (Stream a)
transpose ((c ::: cs) ::: rows) =
    (c ::: map head rows) ::: transpose (cs ::: map tail rows)

-- | The 'subsequences' function returns the stream of all subsequences of the argument.
--   For example:
--
--   > subsequences (0 ::: 1 ::: 2 ::: ...) == [] ::: [0] ::: [1] ::: [0,1] ::: [2] ::: [0,2] ::: [1,2] ::: [0,1,2] ::: ...
subsequences :: Stream a -> Stream [a]
subsequences xs = [] ::: nonEmptySubsequences xs

-- | The 'nonEmptySubsequences' function returns the stream of all subsequences of the argument,
--   except for the empty list.
nonEmptySubsequences :: Stream a -> Stream [a]
nonEmptySubsequences = fold $ \x -> ([x]       :::) . fold (\sub ->
                                    (     sub  :::) .
                                    ((x : sub) :::))

-- Sorry, I couldn't resist writing it in a more point-free (or
-- pointless ;-)) style: The advantage of this version dough is that
-- you can clearly see the three ways a subsequence is build:
--
-- '[x]', 'sub' and 'x:sub'
--
-- This is the version from Data.List which uses more explicit
-- recursion and is point-full:
--
-- nonEmptySubsequences (x:::xs)  =  [x] ::: fold f (nonEmptySubsequences xs)
--     where f sub rest = sub ::: (x : sub) ::: rest

{-| The 'permutations' function returns the stream of all permutations of the argument.
For example:

@
permutations (0 ::: 1 ::: 2 ::: 3 ::: 4 ::: ...) ==
(   (0 ::: 1 ::: 2 ::: 3 ::: 4 ::: ...)
::: (1 ::: 0 ::: 2 ::: 3 ::: 4 ::: ...)
::: (1 ::: 2 ::: 0 ::: 3 ::: 4 ::: ...)
::: (2 ::: 0 ::: 1 ::: 3 ::: 4 ::: ...)
::: (1 ::: 2 ::: 3 ::: 0 ::: 4 ::: ...)
::: (2 ::: 1 ::: 0 ::: 3 ::: 4 ::: ...)
::: (2 ::: 0 ::: 3 ::: 1 ::: 4 ::: ...)
::: (1 ::: 2 ::: 3 ::: 4 ::: 0 ::: ...)
::: (2 ::: 1 ::: 3 ::: 0 ::: 4 ::: ...)
::: (2 ::: 3 ::: 0 ::: 1 ::: 4 ::: ...)
::: ...
)
@

Explaination:

The permutations of @xxs\@(x ::: xs)@ are delivered as follows:

We first deliver @xxs@ because it is a permutation of itself:

@
permutations xxs\@(x ::: xs) = xxs ::: ???
@

Then we calculate the permutations of the tail of @xxs@:

@
permutations xxs\@(x ::: xs) = xxs ::: ??? (permutations xs)
@

Now for each permutation of @xs@ we calculate a new stream of
permutations where we put the @x@ /at every position/ in the
permutation:

@
permutations xxs\@(x ::: xs) = xxs ::: (??? map weave_x) (permutations xs)
    where
       weave_x (p ::: ps) = (p ::: x ::: ps) ::: map (p :::) (weave_x ps)
@

Note that @weave_x@ calculates a new stream of permutations where the
@x@ is put /at every position/ in the given permutation. For
example:

@
weave_x (1 ::: 2 ::: 3 ::: 4 ::: 5 ...) ==
(   (1 ::: x ::: 2 ::: 3 ::: 4 ::: 5 ...)
::: (1 ::: 2 ::: x ::: 3 ::: 4 ::: 5 ...)
::: (1 ::: 2 ::: 3 ::: x ::: 4 ::: 5 ...)
::: (1 ::: 2 ::: 3 ::: 4 ::: x ::: 5 ...)
::: (1 ::: 2 ::: 3 ::: 4 ::: 5 ::: x ...)
...
)
@

Note that we don't put the @x@ at the beginning because we already
delivered that permutation (namely @xxs@)

Now we have a stream of streams of permutations that represent all the
permutations of @xxs@ except @xxs@ itself. The last step is to
/combine/ this stream of streams of permutations into a single stream
of permutations. This is done using the function 'concatDiagonals':

@
permutations xxs\@(x ::: xs) = xxs ::: (concatDiagonals . map weave_x) (permutations xs)
    where
      weave_x (p ::: ps) = (p ::: x ::: ps) ::: map (p :::) (weave_x ps)
@
-}
permutations :: Stream a -> Stream (Stream a)
permutations xxs@(x ::: xs) = xxs ::: (concatDiagonals . map weave_x) (permutations xs)
    where
      weave_x (p ::: ps) = (p ::: x ::: ps) ::: map (p :::) (weave_x ps)

-- | Interleave two Streams @xs@ and @ys@, alternating elements
-- from each list.
-- For example:
--
-- > (x1 ::: x2 ::: ...) `interleave` (y1 ::: y2 ::: ...) == x1 ::: y1 ::: x2 ::: y2 ::: ...
interleave :: Stream a -> Stream a -> Stream a
(x ::: xs) `interleave` ys = x ::: (ys `interleave` xs)

-- | 'scan' yields a stream of successive reduced values.
-- For example:
--
-- > scan f z (x1 ::: x2 ::: ...) == z ::: z `f` x1 ::: (z `f` x1) `f` x2 ::: ...
scan :: (a -> b -> a) -> a -> Stream b -> Stream a
scan f = scan_f
    where
      scan_f z xxs = z ::: case xxs of
                             x ::: xs -> scan_f (f z x) xs

-- | A strict 'scan'. For example:
--
-- > scan (+) 0 (repeat 1) !! 1000000
-- > *** Exception: stack overflow
--
-- > scan' (+) 0 (repeat 1) !! 1000000
-- > 1000000
scan' :: (a -> b -> a) -> a -> Stream b -> Stream a
scan' f = scan'_f
    where
      scan'_f z xxs = z ::: case xxs of
                              x ::: xs -> (scan'_f $! f z x) xs

-- | 'scan1' is a variant of 'scan' that has no starting value argument:
-- For example:
--
-- > scan1 f (x1 ::: x2 ::: ...) == x1 ::: x1 `f` x2 ::: ...
scan1 :: (a -> a -> a) -> Stream a -> Stream a
scan1 f (x ::: xs) = scan f x xs

-- | A strict 'scan1'.
scan1' :: (a -> a -> a) -> Stream a -> Stream a
scan1' f (x ::: xs) = scan' f x xs


--------------------------------------------------------------------------------
-- Diagonals
--------------------------------------------------------------------------------

{-| 'diagonals' @xss@ returns the diagonals of the stream of streams
@xss@ that run from the top to lower left. For example:

@ diagonals
(   (0 :::  1 :::  3 :::  6 ::: ...)
::: (2 :::  4 :::  7 ::: 11 ::: ...)
::: (5 :::  8 ::: 12 ::: 17 ::: ...)
::: (9 ::: 13 ::: 18 ::: 24 ::: ...)
::: ...
) == [0] ::: [1,2] ::: [3,4,5] ::: [6,7,8,9] ::: ...
@
-}
diagonals :: Stream (Stream a) -> Stream [a]
diagonals = fold $ \(x ::: xs) -> ([x] :::) . zipWith (:) xs

-- | 'unDiagonals' is the inverse of 'diagonals'. I.e. it turns a
-- stream of diagonals back into the original stream of streams.
unDiagonals :: Stream [a] -> Stream (Stream a)
unDiagonals = fold zipDiagonal
    where
      zipDiagonal []     rows = rows
      zipDiagonal (d:ds) rows = (d ::: head rows) ::: zipDiagonal ds (tail rows)

-- | 'concatDiagonals' @xss@ concats the 'diagonals' of @xss@.
concatDiagonals :: Stream (Stream a) -> Stream a
concatDiagonals = concat . diagonals

-- | 'unConcatDiagonals' is the inverse of 'concatDiagonals'. I.e. it
-- turns a concatted stream of 'diagonals' back into a stream of
-- diagonals
unConcatDiagonals :: Stream a -> Stream (Stream a)
unConcatDiagonals = unDiagonals . consSplitAt 1
    where
      consSplitAt n xs = let (diagonal, ys) = splitAt n xs
                         in diagonal ::: consSplitAt (n+1) ys


--------------------------------------------------------------------------------
-- Building streams
--------------------------------------------------------------------------------

-- | 'iterate' @f@ @x@ function produces the infinite sequence
-- of repeated applications of @f@ to @x@.
--
-- > iterate f x = x ::: f x ::: f (f x) ::: ...
iterate :: (a -> a) -> a -> Stream a
iterate f = iterate_f
    where
      iterate_f x = x ::: (iterate_f $ f x)

-- | A strict 'iterate'. For example:
--
-- > iterate (+1) 0 !! 1000000
-- > *** Exception: stack overflow
--
-- > iterate' (+1) 0 !! 1000000
-- > 1000000
iterate' :: (a -> a) -> a -> Stream a
iterate' f = iterate'_f
    where
      iterate'_f x = x ::: (iterate'_f $! f x)

-- | 'repeat' @x@ returns a constant stream, where all elements are
-- equal to @x@.
repeat :: a -> Stream a
repeat x = repeat_x
    where
      repeat_x = x ::: repeat_x

-- | 'replicateCont' @n@ @x@ @xs@ returns a stream of @n@ @x@'s followed by
-- the stream @xs@.
replicateCont :: Int -> a -> Stream a -> Stream a
replicateCont n x xs = replicateCont_x_xs n
    where
      replicateCont_x_xs 0 = xs
      replicateCont_x_xs n = x ::: replicateCont_x_xs (n-1)

-- | 'cycle' @xs@ returns the infinite repetition of @xs@.
-- For example:
--
-- > cycle [1,2,3] = 1 ::: 2 ::: 3 ::: 1 ::: 2 ::: 3 ::: ...
cycle :: [a] -> Stream a
cycle xs = cycle_xs
    where
      cycle_xs = foldr (:::) cycle_xs xs

-- | The unfold function is similar to the unfold for lists. Note
-- there is no base case: all streams must be infinite.
unfold :: (b -> (a, b)) -> b -> Stream a
unfold f = unfold_f
    where
      unfold_f y = let (x, z) = f y
                   in x ::: unfold_f z

-- | A strict 'unfold'. For example
--
-- > unfold (\x -> (x, x+1)) 0 !! 1000000
-- > *** Exception: stack overflow
--
-- > unfold' (\x -> (x, x+1)) 0 !! 1000000
-- > 1000000
unfold' :: (b -> (a, b)) -> b -> Stream a
unfold' f = unfold'_f
    where
      unfold'_f y = let (x, z) = f y
                    in x ::: (unfold'_f $! z)


--------------------------------------------------------------------------------
-- Extracting sublists
--------------------------------------------------------------------------------

-- | 'take' @n@ @xs@ returns the first @n@ elements of @xs@.
take :: Int -> Stream a  -> [a]
take = genericTake

{-# SPECIALISE genericTake :: Int -> Stream a -> [a] #-}

-- | The 'genericTake' function is an overloaded version of 'take', which
-- accepts any 'Integral' value as the number of elements to take.
genericTake :: (Integral i) => i -> Stream a -> [a]
genericTake n _ | n <= 0 = []
genericTake n (x ::: xs) = x : genericTake (n - 1) xs

-- | 'drop' @n@ @xs@ drops the first @n@ elements off the front of
-- the sequence @xs@.
drop :: Int -> Stream a -> Stream a
drop = genericDrop

{-# SPECIALISE genericDrop :: Int -> Stream a -> Stream a #-}

-- | The 'genericDrop' function is an overloaded version of 'drop', which
-- accepts any 'Integral' value as the number of elements to drop.
genericDrop :: (Integral i) => i -> Stream a -> Stream a
genericDrop n xs | n <= 0 = xs
genericDrop n xs          = genericDrop (n - 1) (tail xs)

-- | The 'splitAt' function takes an integer @n@ and a stream @xs@
-- and returns a pair consisting of the prefix of @xs@ of length
-- @n@ and the remaining stream immediately following this prefix.
splitAt :: Int -> Stream a -> ([a], Stream a)
splitAt = genericSplitAt

{-# SPECIALISE genericSplitAt :: Int -> Stream a -> ([a], Stream a) #-}

-- | The 'genericSplitAt' function is an overloaded version of 'splitAt', which
-- accepts any 'Integral' value as the position at which to split.
genericSplitAt :: (Integral i) => i -> Stream a -> ([a], Stream a)
genericSplitAt n xs | n <= 0 = ([], xs)
genericSplitAt n (x ::: xs)  = let (prefix, rest) = genericSplitAt (n-1) xs
                               in (x:prefix, rest)

-- | 'takeWhile' @p@ @xs@ returns the longest prefix of the stream
-- @xs@ for which the predicate @p@ holds.
takeWhile :: (a -> Bool) -> Stream a -> [a]
takeWhile p = fold $ \x xs -> if p x then x : xs else []

-- | 'dropWhile' @p@ @xs@ returns the suffix remaining after
-- 'takeWhile' @p@ @xs@.
--
-- /Beware/: this function diverges if every element of @xs@ satisfies
-- @p@, e.g. @dropWhile even (repeat 0)@ will loop.
dropWhile :: (a -> Bool) -> Stream a -> Stream a
dropWhile p = dropWhile_p
    where
      dropWhile_p xxs@(x ::: xs) | p x       = dropWhile_p xs
                                 | otherwise = xxs

-- | 'span' @p@ @xs@ returns the longest prefix of @xs@ that satisfies
-- @p@, together with the remainder of the stream.
span :: (a -> Bool) -> Stream a -> ([a], Stream a)
span p = span_p
    where
      span_p xxs@(x ::: xs) | p x       = first (x:) $ span_p xs
                            | otherwise = ([], xxs)

-- | 'break' @p@ is equivalent to 'span' @(not . p)@.
break :: (a -> Bool) -> Stream a -> ([a], Stream a)
break p = span (not . p)

-- | 'filter' @p@ @xs@, removes any elements from @xs@ that do not satisfy @p@.
--
-- /Beware/: this function may diverge if there is no element of
-- @xs@ that satisfies @p@, e.g. @filter odd (repeat 0)@ will loop.
filter :: (a -> Bool) -> Stream a -> Stream a
filter p = fold $ \x xs -> if p x then x ::: xs else xs

-- | The 'partition' function takes a predicate @p@ and a stream
-- @xs@, and returns a pair of streams. The first stream corresponds
-- to the elements of @xs@ for which @p@ holds; the second stream
-- corresponds to the elements of @xs@ for which @p@ does not hold.
--
-- /Beware/: One of the elements of the tuple may be undefined. For
-- example, @fst (partition even (repeat 0)) == repeat 0@; on the
-- other hand @snd (partition even (repeat 0))@ is undefined.
partition :: (a -> Bool) -> Stream a -> (Stream a, Stream a)
partition p = fold $ \x -> (if p x then first else second) (x :::)

-- | The 'group' function takes a stream and returns a stream of lists such
-- that the concatenation of the result is equal to the argument.  Moreover,
-- each sublist in the result contains only equal elements.  For example,
--
-- > group $ cycle "Mississippi" = "M" ::: "i" ::: "ss" ::: "i" ::: "ss" ::: "i" ::: "pp" ::: "i" ::: "M" ::: "i" ::: ...
--
-- It is a special case of 'groupBy', which allows the programmer to supply
-- their own equality test.
group :: Eq a => Stream a -> Stream [a]
group = groupBy (==)

-- | The 'groupBy' function is the non-overloaded version of 'group'.
groupBy :: (a -> a -> Bool) -> Stream a -> Stream [a]
groupBy eq = groupBy_eq
    where
      groupBy_eq (x ::: xs) = let (ys, zs) = span (eq x) xs
                              in (x : ys) ::: groupBy_eq zs


--------------------------------------------------------------------------------
-- Stream prefixes
--------------------------------------------------------------------------------

-- | The 'isPrefixOf' function returns @True@ if the first argument is
-- a prefix of the second.
isPrefixOf :: Eq a => [a] -> Stream a -> Bool
isPrefixOf []     _          = True
isPrefixOf (p:ps) (x ::: xs) = p == x && isPrefixOf ps xs

-- | The 'stripPrefix' function drops the given prefix from a stream.
-- It returns 'Nothing' if the stream did not start with the prefix
-- given, or 'Just' the stream after the prefix, if it does.
--
-- > stripPrefix "foo" $ cycle "foobar" -> Just "barbarbar..."
-- > stripPrefix "foo" $ cycle "barfoo" -> Nothing
stripPrefix :: Eq a => [a] -> Stream a -> Maybe (Stream a)
stripPrefix []     xs                   = Just xs
stripPrefix (p:ps) (x:::xs) | p == x    = stripPrefix ps xs
                            | otherwise = Nothing


--------------------------------------------------------------------------------
-- Indexing streams
--------------------------------------------------------------------------------

-- | @xs !! n@ returns the element of the stream @xs@ at index
-- @n@. Note that the head of the stream has index 0.
--
-- /Beware/: passing a negative integer as the first argument will cause
-- an error.
(!!) :: Stream a -> Int -> a
(!!) = genericIndex

{-# SPECIALISE genericIndex :: Stream a -> Int -> a #-}

-- | The 'genericIndex' function is an overloaded version of '!!', which
-- accepts any 'Integral' value as the index.
genericIndex :: (Integral i) => Stream a -> i -> a
genericIndex (x ::: xs) n
  | n == 0    = x
  | n > 0     = genericIndex xs (n - 1)
  | otherwise = error "Data.Stream.genericIndex / (!!) negative argument"

-- | The 'elemIndex' function returns the index of the first element
-- in the given stream which is equal (by '==') to the query element,
--
-- /Beware/: 'elemIndex' @x@ @xs@ will diverge if none of the elements
-- of @xs@ equal @x@.
elemIndex :: Eq a => a -> Stream a -> Int
elemIndex x = findIndex (x==)

-- | The 'elemIndices' function extends 'elemIndex', by returning the
-- indices of all elements equal to the query element, in ascending order.
--
-- /Beware/: 'elemIndices' @x@ @xs@ will diverge if none of the
-- elements of @xs@ equal @x@.
elemIndices :: Eq a => a -> Stream a -> Stream Int
elemIndices x = findIndices (x==)

-- | The 'findIndex' function takes a predicate and a stream and returns
-- the index of the first element in the stream satisfying the predicate,
--
-- /Beware/: 'findIndex' @p@ @xs@ will diverge if none of the elements of
-- @xs@ satisfy @p@.
findIndex :: (a -> Bool) -> Stream a -> Int
findIndex p = indexFrom 0
    where
      indexFrom ix (x ::: xs) | p x       = ix
                              | otherwise = (indexFrom $! (ix+1)) xs

-- | The 'findIndices' function extends 'findIndex', by returning the
-- indices of all elements satisfying the predicate, in ascending
-- order.
--
-- /Beware/: 'findIndices' @p@ @xs@ will diverge if none of the elements of
-- @xs@ satisfy @p@.
findIndices :: (a -> Bool) -> Stream a -> Stream Int
findIndices p = indicesFrom 0
    where
      indicesFrom ix (x ::: xs) | p x       = ix ::: ixs
                                | otherwise = ixs
          where
            ixs = (indicesFrom $! (ix+1)) xs


--------------------------------------------------------------------------------
-- Searching streams
--------------------------------------------------------------------------------

-- | 'lookup' @key@ @assocs@ looks up a key in an association stream.
--
--  /Beware/: This function diverges if @key@ isn't in @assocs@.
lookup :: Eq k => k -> Stream (k, a) -> a
lookup key = fold $ \(k, v1) v2 -> if key == k then v1 else v2


--------------------------------------------------------------------------------
-- Zipping and unzipping streams
--------------------------------------------------------------------------------

-- TODO: Use some Template Haskell to abstract the following:

-- | The 'zip' function takes two streams and returns a list of
-- corresponding pairs.
zip :: Stream a -> Stream b -> Stream (a,b)
zip = zipWith (,)

zip3 :: Stream a -> Stream b -> Stream c -> Stream (a,b,c)
zip3 = zipWith3 (,,)

zip4 :: Stream a -> Stream b -> Stream c -> Stream d -> Stream (a,b,c,d)
zip4 = zipWith4 (,,,)

zip5 :: Stream a -> Stream b -> Stream c -> Stream d -> Stream e -> Stream (a,b,c,d,e)
zip5 = zipWith5 (,,,,)

zip6 :: Stream a -> Stream b -> Stream c -> Stream d -> Stream e -> Stream f -> Stream (a,b,c,d,e,f)
zip6 = zipWith6 (,,,,,)

zip7 :: Stream a -> Stream b -> Stream c -> Stream d -> Stream e -> Stream f -> Stream g -> Stream (a,b,c,d,e,f,g)
zip7 = zipWith7 (,,,,,,)

-- | The 'zipWith' function generalizes 'zip'. Rather than tupling
-- the functions, the elements are combined using the function
-- passed as the first argument to 'zipWith'.
zipWith :: (a -> b -> c) -> Stream a -> Stream b -> Stream c
zipWith h = zipWith_h
    where
      zipWith_h (a ::: as) (b ::: bs) = h a b ::: zipWith_h as bs

zipWith3 :: (a -> b -> c -> d) -> Stream a -> Stream b -> Stream c -> Stream d
zipWith3 h = zipWith3_h
    where
      zipWith3_h (a ::: as) (b ::: bs) (c ::: cs) = h a b c ::: zipWith3_h as bs cs

zipWith4 :: (a -> b -> c -> d -> e) -> Stream a -> Stream b -> Stream c -> Stream d -> Stream e
zipWith4 h = zipWith4_h
    where
      zipWith4_h (a ::: as) (b ::: bs) (c ::: cs) (d ::: ds) = h a b c d ::: zipWith4_h as bs cs ds

zipWith5 :: (a -> b -> c -> d -> e -> f) -> Stream a -> Stream b -> Stream c -> Stream d -> Stream e -> Stream f
zipWith5 h = zipWith5_h
    where
      zipWith5_h (a ::: as) (b ::: bs) (c ::: cs) (d ::: ds) (e ::: es) = h a b c d e ::: zipWith5_h as bs cs ds es

zipWith6 :: (a -> b -> c -> d -> e -> f -> g) -> Stream a -> Stream b -> Stream c -> Stream d -> Stream e -> Stream f -> Stream g
zipWith6 h = zipWith6_h
    where
      zipWith6_h (a ::: as) (b ::: bs) (c ::: cs) (d ::: ds) (e ::: es) (f ::: fs) = h a b c d e f ::: zipWith6_h as bs cs ds es fs

zipWith7 :: (a -> b -> c -> d -> e -> f -> g -> h) -> Stream a -> Stream b -> Stream c -> Stream d -> Stream e -> Stream f -> Stream g -> Stream h
zipWith7 h = zipWith7_h
    where
      zipWith7_h (a ::: as) (b ::: bs) (c ::: cs) (d ::: ds) (e ::: es) (f ::: fs) (g ::: gs) = h a b c d e f g ::: zipWith7_h as bs cs ds es fs gs

-- | The 'unzip' function is the inverse of the 'zip' function.
unzip :: Stream (a, b) -> (Stream a, Stream b)
unzip = fold $ \(a, b) ~(as, bs) -> (a ::: as, b ::: bs)

unzip3 :: Stream (a, b, c) -> (Stream a, Stream b, Stream c)
unzip3 = fold $ \(a, b, c) ~(as, bs, cs) -> (a ::: as, b ::: bs, c ::: cs)

unzip4 :: Stream (a, b, c, d) -> (Stream a, Stream b, Stream c, Stream d)
unzip4 = fold $ \(a, b, c, d) ~(as, bs, cs, ds) -> (a ::: as, b ::: bs, c ::: cs, d ::: ds)

unzip5 :: Stream (a, b, c, d, e) -> (Stream a, Stream b, Stream c, Stream d, Stream e)
unzip5 = fold $ \(a, b, c, d, e) ~(as, bs, cs, ds, es) -> (a ::: as, b ::: bs, c ::: cs, d ::: ds, e ::: es)

unzip6 :: Stream (a, b, c, d, e, f) -> (Stream a, Stream b, Stream c, Stream d, Stream e, Stream f)
unzip6 = fold $ \(a, b, c, d, e, f) ~(as, bs, cs, ds, es, fs) -> (a ::: as, b ::: bs, c ::: cs, d ::: ds, e ::: es, f ::: fs)

unzip7 :: Stream (a, b, c, d, e, f, g) -> (Stream a, Stream b, Stream c, Stream d, Stream e, Stream f, Stream g)
unzip7 = fold $ \(a, b, c, d, e, f, g) ~(as, bs, cs, ds, es, fs, gs) -> (a ::: as, b ::: bs, c ::: cs, d ::: ds, e ::: es, f ::: fs, g ::: gs)


--------------------------------------------------------------------------------
-- Functions on streams of characters
--------------------------------------------------------------------------------

-- | The 'words' function breaks a stream of characters into a
-- stream of words, which were delimited by white space.
--
-- Note that @words = breaks isSpace@
--
-- /Beware/: if the stream of characters @xs@ does not contain white
-- space, accessing the tail of @words xs@ will loop.
words :: Stream Char -> Stream String
words = breaks isSpace

-- | The 'unwords' function is an inverse operation to 'words'. It
-- joins words with separating spaces.
unwords :: Stream String -> Stream Char
unwords = unbreak ' '

-- | The 'lines' function breaks a stream of characters into a list
-- of strings at newline characters. The resulting strings do not
-- contain newlines.
--
-- Note that @lines = breaks (== '\n')@
--
-- /Beware/: if the stream of characters @xs@ does not contain
-- newline characters, accessing the tail of @lines xs@ will loop.
lines :: Stream Char -> Stream String
lines = breaks (== '\n')

-- | The 'unlines' function is an inverse operation to 'lines'. It
-- joins lines, after appending a terminating newline to each.
unlines :: Stream String -> Stream Char
unlines = unbreak '\n'

-- | 'breaks' @p@ @xs@ breaks the stream @xs@ into lists where the
-- elements do not satisfy @p@.
--
-- > breaks isSpace (listToStream (Prelude.cycle "streams that contain spaces ")) ==
-- > "streams" ::: "that" ::: "contain" ::: "spaces" ::: "streams" ::: ...
--
-- Also see the specialisations: 'words' and 'lines'.
--
-- /Beware/: 'breaks' @p@ @xs@ will diverge if all elements of @xs@
-- satify @p@.
breaks :: (a -> Bool) -> Stream a -> Stream [a]
breaks p = breaks_p
    where
      breaks_p (x ::: xs)
          | p x       = breaks_p xs
          | otherwise = let (ys, zs) = break p xs
                        in (x:ys) ::: breaks_p (tail zs)

-- | 'unbreak' is an inverse operation to 'breaks'. Also see the
-- specialisations: 'unwords' and 'unlines'. Also see the similar:
-- 'intercalate'
unbreak :: a -> Stream [a] -> Stream a
unbreak y = fold $ \xs -> fromListCont xs . (y :::)


--------------------------------------------------------------------------------
-- Converting to and from an infinite list
--------------------------------------------------------------------------------

-- | 'toList' @s@ converts the stream @s@ into an /infinite/ list.
toList :: Stream a -> [a]
toList = fold (:)

-- | 'fromList' @xs@ converts the /infinite/ list @xs@ to a
-- stream.
--
-- /Beware/: Passing a /finite/ list, will cause an error.
fromList :: [a] -> Stream a
fromList = foldr (:::) (error "Data.Stream.fromList applied to a finite list")

-- | 'fromListCont' @xs@ @s@ converts the list @xs@ into a
-- stream. When @xs@ runs empty the stream @s@ is returned.
fromListCont :: [a] -> Stream a -> Stream a
fromListCont xs s = foldr (:::) s xs


--------------------------------------------------------------------------------
-- Set operations
--------------------------------------------------------------------------------

-- | The 'nub' function removes duplicate elements from a stream.
-- In particular, it keeps only the first occurrence of each element.
-- (The name 'nub' means \`essence\'.)
-- It is a special case of 'nubBy', which allows the programmer to supply
-- their own equality test.
--
-- /Beware/: the 'tail' of 'nub' applied to a stream of equal elements
-- will diverge. For example:
--
-- > nub $ repeat 'a'
--
-- results in: @'a' ::: _|_@
nub :: (Eq a) => Stream a -> Stream a
nub = nubBy (==)

-- | The 'nubBy' function behaves just like 'nub', except it uses a
-- user-supplied equality predicate instead of the overloaded '=='
-- function.
nubBy :: (a -> a -> Bool) -> Stream a -> Stream a
nubBy eq (x ::: xs) = x ::: go [x] xs
    where
      go seen (x ::: xs)
          | elemBy eq x seen = go seen xs
          | otherwise        = x ::: go (x:seen) xs

elemBy :: (a -> a -> Bool) -> a -> [a] -> Bool
elemBy eq x = foldr ((||) . eq x) False

-- | 'delete' @x@ removes the first occurrence of @x@ from its stream argument.
-- For example,
--
-- > delete 'a' $ cycle "banana" == "bnanabanana..."
--
-- It is a special case of 'deleteBy', which allows the programmer to
-- supply their own equality test.
delete :: (Eq a) => a -> Stream a -> Stream a
delete =  deleteBy (==)

-- | The 'deleteBy' function behaves like 'delete', but takes a
-- user-supplied equality predicate.
deleteBy :: (a -> a -> Bool) -> a -> Stream a -> Stream a
deleteBy eq x = deleteBy_eq_x
    where
      deleteBy_eq_x (y ::: ys)
          | eq x y    = ys
          | otherwise = y ::: deleteBy_eq_x ys


--------------------------------------------------------------------------------
-- Ordered streams
--------------------------------------------------------------------------------

-- | The 'insert' function takes an element and a stream and inserts the
-- element into the stream at the last position where it is still less
-- than or equal to the next element.  In particular, if the stream
-- is sorted before the call, the result will also be sorted.
-- It is a special case of 'insertBy', which allows the programmer to
-- supply their own comparison function.
--
-- > insert 3 $ iterate (+1) 0 == 0 ::: 1 ::: 2 ::: 3 ::: 3 ::: 4 ::: 5 ::: ...
insert :: Ord a => a -> Stream a -> Stream a
insert = insertBy compare

-- | The non-overloaded version of 'insert'.
insertBy :: (a -> a -> Ordering) -> a -> Stream a -> Stream a
insertBy cmp x = insertBy_cmp_x
    where
      insertBy_cmp_x yys@(y:::ys) = case cmp x y of
                                      GT -> y ::: insertBy_cmp_x ys
                                      _  -> x ::: yys

-- | Merge two sorted streams into a single, sorted whole.
merge :: Ord a => Stream a -> Stream a -> Stream a
merge = mergeBy compare

-- | The non-overloaded version of 'merge'.
mergeBy :: (a -> a -> Ordering) -> Stream a -> Stream a -> Stream a
mergeBy cmp = mergeBy_cmp
    where
      mergeBy_cmp xxs@(x ::: xs)
                  yys@(y ::: ys) = case x `cmp` y of
                                     GT -> y ::: mergeBy_cmp xxs ys
                                     _  -> x ::: mergeBy_cmp xs  yys


-- The End ---------------------------------------------------------------------
